@@ -445,7 +445,10 @@ class TypeConverter(object):
         count = 1
         end = start
         while True:
-            token = tokens[end]
+            try:
+                token = tokens[end]
+            except IndexError:
+                raise ParseError(tokens)
             end += 1
             if token.name == '<':
                 count += 1
@@ -592,11 +595,21 @@ class TypeConverter(object):
         default = []
 
         def add_parameter():
+            if not type_modifiers:
+                return
             if default:
                 del default[0]  # Remove flag.
             end = type_modifiers[-1].end
-            name, type_name, templated_types, modifiers, _, __ = \
-                self.declaration_to_parts(type_modifiers, True)
+
+            needs_name_removed = not (
+                len(type_modifiers) == 1 or
+                (len(type_modifiers) == 2 and
+                 type_modifiers[0].name == 'const'))
+
+            (name, type_name, templated_types, modifiers,
+             _, __) = self.declaration_to_parts(type_modifiers,
+                                                needs_name_removed)
+
             parameter_type = Type(first_token.start, first_token.end,
                                   type_name, templated_types, modifiers,
                                   reference, pointer, array)
@@ -762,14 +775,16 @@ class AstBuilder(object):
             if last_token.name == '(':
                 # If there is an assignment before the paren,
                 # this is an expression, not a method.
-                if (
-                    temp_tokens[-1].name == '=' and
-                    temp_tokens[-2].name != 'operator'
-                ):
-                    new_temp = self._get_tokens_up_to(tokenize.SYNTAX, ';')
-                    temp_tokens.append(last_token)
-                    temp_tokens.extend(new_temp)
-                    last_token = tokenize.Token(tokenize.SYNTAX, ';', 0, 0)
+                for i, elt in reversed(list(enumerate(temp_tokens))):
+                    if (
+                        elt.name == '=' and
+                        temp_tokens[i - 1].name != 'operator'
+                    ):
+                        new_temp = self._get_tokens_up_to(tokenize.SYNTAX, ';')
+                        temp_tokens.append(last_token)
+                        temp_tokens.extend(new_temp)
+                        last_token = tokenize.Token(tokenize.SYNTAX, ';', 0, 0)
+                        break
 
             if last_token.name == '[':
                 # Handle array, this isn't a method, unless it's an operator.
@@ -980,12 +995,11 @@ class AstBuilder(object):
             assert token.name == '(', token
 
         name = return_type_and_name.pop()
-        if (
-            len(return_type_and_name) > 2 and
-            return_type_and_name[-1].name == 'operator'
-        ):
+        if (len(return_type_and_name) > 1 and
+            (return_type_and_name[-1].name == 'operator' or
+             return_type_and_name[-1].name == '~')):
             op = return_type_and_name.pop()
-            name = tokenize.Token(tokenize.NAME, 'operator' + name.name,
+            name = tokenize.Token(tokenize.NAME, op.name + name.name,
                                   op.start, name.end)
         # Handle templatized ctors.
         elif name.name == '>':
@@ -1027,31 +1041,38 @@ class AstBuilder(object):
             parameters = list(self._get_parameters())
             del parameters[-1]          # Remove trailing ')'.
 
-        token = self._get_next_token()
-        while token.token_type == tokenize.NAME:
-            modifier_token = token
+        try:
             token = self._get_next_token()
-            if modifier_token.name == 'const':
+        except StopIteration:
+            token = tokenize.Token(tokenize.SYNTAX, ';', 0, 0)
+
+        while token.token_type == tokenize.NAME:
+            if token.name == 'const':
                 modifiers |= FUNCTION_CONST
-            elif modifier_token.name == '__attribute__':
+                token = self._get_next_token()
+            elif token.name == '__attribute__':
                 # TODO(nnorwitz): handle more __attribute__ details.
                 modifiers |= FUNCTION_ATTRIBUTE
+                token = self._get_next_token()
                 assert token.name == '(', token
                 # Consume everything between the (parens).
                 list(self._get_matching_char('(', ')'))
                 token = self._get_next_token()
-            elif modifier_token.name == 'throw':
+            elif token.name == 'throw':
                 modifiers |= FUNCTION_THROW
+                token = self._get_next_token()
                 assert token.name == '(', token
                 # Consume everything between the (parens).
                 list(self._get_matching_char('(', ')'))
                 token = self._get_next_token()
-            elif modifier_token.name == modifier_token.name.upper():
+            elif token.name == token.name.upper():
                 # HACK(nnorwitz):  assume that all upper-case names
                 # are some macro we aren't expanding.
                 modifiers |= FUNCTION_UNKNOWN_ANNOTATION
+                token = self._get_next_token()
             else:
-                self.handle_error('unexpected token', modifier_token)
+                self._add_back_token(token)
+                token = tokenize.Token(tokenize.SYNTAX, ';', 0, 0)
 
         if token.token_type != tokenize.SYNTAX:
             raise ParseError(token)
@@ -1074,8 +1095,7 @@ class AstBuilder(object):
                 del function_parameters[-1]  # Remove trailing ')'.
                 # TODO(nnorwitz): store the function_parameters.
                 token = self._get_next_token()
-                assert token.token_type == tokenize.SYNTAX, token
-                if token.name != ';':
+                if token.token_type != tokenize.SYNTAX or token.name != ';':
                     raise ParseError(token)
                 return self._create_variable(indices, name.name, indices.name,
                                              modifiers, '')
@@ -1115,7 +1135,7 @@ class AstBuilder(object):
                 raise ParseError((token, return_type_and_name, parameters))
 
         # Looks like we got a method, not a function.
-        if len(return_type) > 2 and return_type[-1].name == '::':
+        if len(return_type) > 1 and return_type[-1].name == '::':
             return_type, in_class = \
                 self._get_return_type_and_class_name(return_type)
             return Method(indices.start, indices.end, name.name, in_class,
@@ -1236,16 +1256,15 @@ class AstBuilder(object):
         assert token.token_type == tokenize.NAME, token
         return self._create_variable(token, token.name, name, [], '')
 
-    def handle_struct(self):
-        # Special case the handling typedef/aliasing of structs here.
-        # It would be a pain to handle in the class code.
+    def _handle_class_and_struct(self, class_type, class_str, visibility):
+        # Special case the handling typedef/aliasing of classes/structs here.
         name_tokens, var_token = self.get_name()
         if name_tokens:
             # Forward declaration.
             if var_token.name == ';':
-                return Struct(name_tokens[0].start, name_tokens[0].end,
-                              name_tokens[0].name, None, None, None,
-                              self.namespace_stack)
+                return class_type(name_tokens[0].start, name_tokens[0].end,
+                                  name_tokens[0].name, None, None, None,
+                                  self.namespace_stack)
 
             next_token = self._get_next_token()
             is_syntax = (var_token.token_type == tokenize.SYNTAX and
@@ -1257,18 +1276,18 @@ class AstBuilder(object):
                 variable = next_token
                 temp = self._get_next_token()
                 if temp.token_type == tokenize.SYNTAX and temp.name == '(':
-                    # Handle methods declared to return a struct.
+                    # Handle methods declared to return a class/struct.
                     t0 = name_tokens[0]
-                    struct = tokenize.Token(tokenize.NAME, 'struct',
-                                            t0.start - 7, t0.start - 2)
-                    type_and_name = [struct]
+                    token = tokenize.Token(tokenize.NAME, class_str,
+                                           t0.start - 7, t0.start - 2)
+                    type_and_name = [token]
                     type_and_name.extend(name_tokens)
                     type_and_name.extend((var_token, next_token))
                     return self._get_method(type_and_name, 0, None, False)
                 if temp.name != ';':
                     raise ParseError((temp, name_tokens, var_token))
             if is_syntax or (is_variable and not self._handling_typedef):
-                modifiers = ['struct']
+                modifiers = [class_str]
                 type_name = ''.join([t.name for t in name_tokens])
                 position = name_tokens[0]
                 return self._create_variable(
@@ -1278,7 +1297,15 @@ class AstBuilder(object):
             self._add_back_tokens(name_tokens)
         else:
             self._add_back_token(var_token)
-        return self._get_class(Struct, VISIBILITY_PUBLIC, None)
+        return self._get_class(class_type, visibility, None)
+
+    def handle_class(self):
+        return self._handle_class_and_struct(Class, 'class',
+                                             VISIBILITY_PRIVATE)
+
+    def handle_struct(self):
+        return self._handle_class_and_struct(Struct, 'struct',
+                                             VISIBILITY_PUBLIC)
 
     def handle_union(self):
         return self._get_nested_type(Union)
@@ -1287,8 +1314,6 @@ class AstBuilder(object):
         return self._get_nested_type(Enum)
 
     def handle_auto(self):
-        # TODO(nnorwitz): warn about using auto?  Probably not since it
-        # will be reclaimed and useful for C++0x.
         pass
 
     def handle_register(self):
@@ -1476,9 +1501,6 @@ class AstBuilder(object):
     def handle_asm(self):
         pass  # Not needed yet.
 
-    def handle_class(self):
-        return self._get_class(Class, VISIBILITY_PRIVATE, None)
-
     def _get_bases(self):
         # Get base classes.
         bases = []
