@@ -52,6 +52,9 @@ __author__ = 'nnorwitz@google.com (Neal Norwitz)'
 # file.
 PRIMARY_HEADER_EXTENSION = '.h'
 
+HEADER_EXTENSIONS = frozenset(['.h', '.hpp', '.h++', '.hxx'])
+CPP_EXTENSIONS = frozenset(['.cc', '.cpp', '.c++', '.cxx'])
+
 # These enumerations are used to determine how an symbol/#include file is used.
 UNUSED = 0
 USES_REFERENCE = 1
@@ -74,14 +77,14 @@ class Module(object):
         return dict([(n.name, n) for n in self.ast_list if n.is_exportable()])
 
 
-def _is_header_file(filename):
+def is_header_file(filename):
     _, ext = os.path.splitext(filename)
-    return ext.lower() in ('.h', '.hpp', '.h++', '.hxx')
+    return ext.lower() in HEADER_EXTENSIONS
 
 
-def _is_cpp_file(filename):
+def is_cpp_file(filename):
     _, ext = os.path.splitext(filename)
-    return ext.lower() in ('.c', '.cc', '.cpp', '.c++', '.cxx')
+    return ext.lower() in CPP_EXTENSIONS
 
 
 class WarningHunter(object):
@@ -120,12 +123,10 @@ class WarningHunter(object):
                 print('%s:%d: %s' % (filename, line_num, msg))
 
     def find_warnings(self):
-        if _is_header_file(self.filename):
+        if is_header_file(self.filename):
             self._find_header_warnings()
-        elif _is_cpp_file(self.filename):
+        elif is_cpp_file(self.filename):
             self._find_source_warnings()
-        else:
-            print('Unknown filetype for: %s' % self.filename)
 
     def _update_symbol_table(self, module):
         for name, node in module.public_symbols.items():
@@ -189,7 +190,7 @@ class WarningHunter(object):
         files_seen = {}
         for filename, (node, module) in included_files.items():
             normalized_filename = module.normalized_filename
-            if _is_cpp_file(filename):
+            if is_cpp_file(filename):
                 msg = 'should not #include C++ source file: %s' % filename
                 self._add_warning(msg, node)
             if normalized_filename == self.normalized_filename:
@@ -205,12 +206,12 @@ class WarningHunter(object):
     def _verify_include_files_used(self, file_uses, included_files):
         """Find all #include files that are unnecessary."""
         for include_file, use in file_uses.items():
-            if use != USES_DECLARATION:
+            if not use & USES_DECLARATION:
                 node, module = included_files[include_file]
                 if module.ast_list is not None:
                     msg = module.filename + ' does not need to be #included'
-                    if use == USES_REFERENCE:
-                        msg += '. Use references instead'
+                    if use & USES_REFERENCE:
+                        msg += '. Use a forward declaration instead'
                     self._add_warning(msg, node)
 
     def _verify_forward_declarations_used(self, forward_declarations,
@@ -234,13 +235,22 @@ class WarningHunter(object):
         symbol_table = self.symbol_table
 
         def _add_reference(name, namespace):
-            if name in decl_uses:
-                decl_uses[name] |= USES_REFERENCE
-            elif not None in namespace:
+            if not name in decl_uses and namespace and not None in namespace:
                 # TODO(nnorwitz): make less hacky, do real name lookup.
                 name = '::'.join(namespace) + '::' + name
-                if name in decl_uses:
-                    decl_uses[name] |= USES_REFERENCE
+            if name in decl_uses:
+                decl_uses[name] |= USES_REFERENCE
+            else:
+                try:
+                    file_use_node = symbol_table.lookup_symbol(name, namespace)
+                except symbols.Error:
+                    return
+                name = file_use_node[1].normalized_filename
+                if name in file_uses:
+                    if isinstance(file_use_node[0], ast.Typedef):
+                        file_uses[name] |= USES_DECLARATION
+                    else:
+                        file_uses[name] |= USES_REFERENCE
 
         def _add_use(name, namespace):
             if isinstance(name, list):
@@ -258,6 +268,8 @@ class WarningHunter(object):
                 # exception should not happen...unless the code relies
                 # on another header for proper compilation.
                 # Store the use since we might really need to #include it.
+                if namespace and not None in namespace and not '::' in name:
+                    name = '::'.join(namespace) + '::' + name
                 file_uses[name] = file_uses.get(name, 0) | USES_DECLARATION
                 return
             if not file_use_node:
@@ -356,8 +368,6 @@ class WarningHunter(object):
                     if node.body:
                         ast_seq.append(node.body)
                     _add_template_use('', node.bases, node.namespace)
-                elif isinstance(node, ast.Struct) and node.body is not None:
-                    pass  # TODO(nnorwitz): impl
                 elif isinstance(node, ast.Union) and node.fields:
                     pass  # TODO(nnorwitz): impl
 
@@ -365,8 +375,8 @@ class WarningHunter(object):
 
     def _find_unused_warnings(self):
         included_files, forward_declarations = self._read_and_parse_includes()
-        file_uses, decl_uses = \
-            self._determine_uses(included_files, forward_declarations)
+        file_uses, decl_uses = self._determine_uses(included_files,
+                                                    forward_declarations)
         self._verify_includes(included_files)
         self._verify_include_files_used(file_uses, included_files)
         self._verify_forward_declarations_used(forward_declarations, decl_uses,
